#!/usr/bin/env python
# -*- coding: utf-8 -*-
from roller_constant import ForWidget, OptionKey, SessionKey
from roller_box import RollerBox
from roller_button import RollerButton
import gtk


class GroupFormatList(gtk.TreeView):
    """
    This is widget group with TreeView and TreeView manipulation Buttons.

    Use the GroupFormatList class template to
    hook feedback from the GroupFormatList:
        on_format_list_change: function
            Call when the format list changes.

        create_format: function
            Call when the New Format button is activated.

        edit_format: function
            Call when the Edit Format button is activated.

        delete_format: function
            Call when the Delete Format button is activated.
    """
    RESPOND_KEY = 'Return', 'Delete', 'space'

    def __init__(self, g, creator):
        """
        Adds the format list and Buttons to a group.

        g: GTK container
            the top-most container

        d: dict
            widget dict

        creator: class instance
            the one that created the format list
        """
        w = ForWidget.MARGIN
        self.creator = creator
        self.key = SessionKey.FORMAT_LIST
        self.buttons = []
        self.items = []
        g1 = RollerBox(ForWidget.HBOX)

        # Format Stack TreeView:
        self.store = gtk.ListStore(str, str, str)

        gtk.TreeView.__init__(
                self,
                model=self.store
            )

        self.wig = self
        self.set_headers_visible = 0

        # Create column:
        col = gtk.TreeViewColumn(
                "Format Stack:",
                gtk.CellRendererText(),
                text=0,
                foreground=1,
                background=2
            )

        col.set_min_width(120)
        col.set_sizing(gtk.TREE_VIEW_COLUMN_AUTOSIZE)

        # Add column to TreeView:
        self.append_column(col)

        # When a row is selected, it emits a signal:
        self.get_selection().connect('changed', self.selected)
        self.connect('key_press_event', self.on_key_press)
        g1.alignment.set_padding(w / 2, w, w, w)

        g2 = RollerBox(
                ForWidget.VBUTTON_BOX,
                align=(0, 0, 0, 1),
                padding=(0, 0, ForWidget.MARGIN / 2, 0)
            )

        a = zip(
                (
                    "New Format",
                    "Edit Format",
                    "Del Format"),
                (
                    self._create_format,
                    self._edit_format,
                    self._del_format
                ),
                (0, 1, 1)
            )

        for n, p, b in a:
            g3 = GroupFormatList.CoOpButton(n, p, b)
            g2.add(g3.alignment)
            self.buttons.append(g3)

        hbox = gtk.HBox()
        g4 = self._move_up_button = GroupFormatList.CoOpButton(
            "↑", self.move_sel_up, 2)

        g5 = self._move_down_button = GroupFormatList.CoOpButton(
            "↓", self.move_sel_down, 2)

        g6 = gtk.ScrolledWindow()

        hbox.add(g4.alignment)
        hbox.add(g5.alignment)
        g2.add(hbox)
        [self.buttons.append(i) for i in (g4, g5)]
        g6.set_policy(gtk.POLICY_NEVER, gtk.POLICY_AUTOMATIC)
        g6.add_with_viewport(self)
        g1.add(g6)
        g1.add(g2.alignment)
        g.add(g1.alignment)
        self.verify_buttons()

    def _create_format(self, *_):
        """
        Create a new format.

        Enumerates the ‟Format” name until
        a unique identifier is created.
        """
        q = self.items
        n = "Format"
        a = len(q) + 1
        go = 1

        while go:
            go = 0
            n1 = n + " " + str(a)
            for i in q:
                if i[0] == n1:
                    a += 1
                    go = 1
                    break

        self.store.append([n1, '#000000', '#00FEFF'])
        q.append((n1, len(q)))
        self.verify_buttons()
        self.select_item(len(q) - 1)
        self.creator.create_format(n1)

    def _del_format(self, *_):
        """Delete a row in the TreeView."""
        x = self.get_sel_x()

        self.items.pop(x)

        # model, iter:
        a, b = self.wig.get_selection().get_selected()

        a.remove(b)
        self.creator.delete_format(x)
        self.populate_treeview(self.items)

    def _edit_format(self, *_):
        """Open the UIFormat window with the selected format."""
        self.creator.edit_format()

    def get_sel_x(self):
        """Return the row index of the selected item."""
        return self.wig.get_selection().get_selected_rows()[1][0][0]

    def get_sel_text(self):
        """Return the selected format name."""
        return self.items[self.get_sel_x()][0]

    def get_value(self):
        """
        Return a list of strings displayed in the list.

        Is part of a UI widget template.
        """
        return self.items

    def populate_treeview(self, q):
        """
        Populate the TreeView from a list.
        The previous indexed-list is replaced.

        q: list of tuples (name, index)
        """
        self.store.clear()

        self.items = []
        for x, b in enumerate(q):
            self.store.append([b[0], '#000000', '#00FEFF'])
            self.items.append((b[0], x))

    def move_sel_down(self, *_):
        u"""
        Move selections down one line in the TreeView.

        If ‟x” selection is at the bottom of the list,
        then the item rotates to the top of the list.
        """
        x = x1 = self.get_sel_x()
        a = len(self.items)
        x += 1

        if x == a:
            x = 0
        self.update_treeview(x, x1)
        self._move_down_button.wig.grab_focus()

    def move_sel_up(self, *_):
        u"""
        Move selections up one line in the TreeView.

        If ‟x” selection is at the top of the list, then
        the item rotates to the bottom of the list.
        """
        x = x1 = self.get_sel_x()
        a = len(self.items)
        x -= 1

        if x < 0:
            x = a - 1
        self.update_treeview(x, x1)
        self._move_up_button.wig.grab_focus()

    def on_key_press(self, _, a):
        """
        Use to catch the delete and return keys.

        a: key-press event
        """
        n = gtk.gdk.keyval_name(a.keyval)
        if len(self.items):
            if n in GroupFormatList.RESPOND_KEY:
                x = GroupFormatList.RESPOND_KEY.index(n)

                (self._edit_format, self._del_format, self.selected)[x]()
                return 1

    def rename(self, x, n):
        """
        Rename a format.

        x: the current list entry index
        n: name of format
        """
        self.items[x] = n, self.items[x][1]
        self.populate_treeview(self.items)

    def selected(self, *_):
        """Verify the buttons with the current selection state."""
        self.verify_buttons()

    def select_item(self, x):
        """
        Select, sort of, a TreeView item.

        x: row index in the TreeView
        """
        self.wig.set_cursor(x)

    def set_indices(self):
        u"""
        Update the ‟self.items” indices.

        Call after the Session
        dictionary's format list is re-ordered.
        """
        for i in range(len(self.items)):
            self.items[i] = self.items[i][0], i

    def set_value(self, q):
        """
        Load the format list from a session dictionary's list of formats.

        Is part of a UI widget template.

        q: list of formats from the session dict
        """
        b = []
        for d in q:
            b.append((d[OptionKey.NAME], 0))
        self.populate_treeview(b)

    def update_treeview(self, x, x1):
        """
        Update the list after a change.

        x: new position (int)
        x1: old position
        """
        q = self.items[x1]

        self.items.pop(x1)
        self.items.insert(x, q)
        self.creator.on_format_list_change()
        self.populate_treeview(self.items)
        self.select_item(x)

    def verify_buttons(self):
        """Verify Buttons that are dependent on the list quantity."""
        _, b = self.wig.get_selection().get_selected()

        # The move and delete Buttons are dependent
        # on their selection state and the list size:
        for g in self.buttons[1:]:
            if b and len(self.items) >= g.need:
                g.enable()

            else:
                g.disable()

    class CoOpButton(RollerButton):
        """
        Has a variable, ‟self.need”.
        ‟self.need” is the number of items in a list
        needed for the Button to be valid.
        """

        def __init__(self, text, callback, need):
            """
            text: string
                Button Label

            callback: function
                callback on action

            needed: int
                the number of items needed for this Button to be valid
            """
            self.need = need
            RollerButton.__init__(self, text, callback)
